@@ -7,10 +7,20 @@ use rr_util::{
 };
 use uuid::uuid;
 
-use crate::{CachedCircuitInfo, CircuitFlags, CircuitNode, ConstructError, NamedAxes, Result};
+use crate::{
+    CachedCircuitInfo, CircuitFlags, CircuitNode, CircuitNodeAutoName, ConstructError, NamedAxes,
+    Result,
+};
 
 pub trait CircuitNodeInit {
-    fn init_info_impl(self) -> Result<Self>
+    fn initial_init_info_impl(self) -> Result<Self>
+    where
+        Self: Sized,
+    {
+        self.init_info_impl(true)
+    }
+
+    fn init_info_impl(self, is_initial: bool) -> Result<Self>
     where
         Self: Sized;
 
@@ -47,7 +57,6 @@ pub trait CircuitNodeComputeInfoImpl: CircuitNode {
 
     fn compute_device_dtype(&self) -> Result<TorchDeviceDtypeOp> {
         if let Some(z) = self.device_dtype_override() {
-            println!("override device dtype");
             return z;
         }
         self.children()
@@ -124,26 +133,61 @@ impl<T: CircuitNodeHashItems + CircuitNode> CircuitNodeHashWithChildren for T {
     }
 }
 
+pub trait CircuitNodeMaybeAutoName {
+    fn maybe_auto_name(&self, name: Option<String>) -> Option<String> {
+        name
+    }
+}
+impl<T: CircuitNodeAutoName> CircuitNodeMaybeAutoName for T {
+    fn maybe_auto_name(&self, name: Option<String>) -> Option<String> {
+        name.or_else(|| {
+            self.info()
+                .use_autoname()
+                .then(|| self.auto_name())
+                .flatten()
+        })
+    }
+}
+
 impl<T> CircuitNodeInit for T
 where
-    T: CircuitNodeHashWithChildren + CircuitNodePrivate + CircuitNodeSetNonHashInfo + CircuitNode,
+    T: CircuitNodeHashWithChildren
+        + CircuitNodePrivate
+        + CircuitNodeSetNonHashInfo
+        + CircuitNode
+        + CircuitNodeMaybeAutoName
+        + Sized,
 {
-    fn init_info_impl(mut self) -> Result<Self>
-    where
-        Self: Sized,
-    {
+    fn init_info_impl(mut self, is_initial: bool) -> Result<Self> {
         let mut hasher = blake3::Hasher::new();
         self.compute_hash_non_name(&mut hasher);
         hasher.update(&self.node_type_uuid());
-        hasher.update(self.name().unwrap_or("").as_bytes());
-        hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes()); // delimit name with uuid
 
+        // note that we this might be set by a prior construction
+        // Because we can be called by (e.g.) rename_impl or update_info_impl
+        let prior_autoname = self.info().use_autoname();
         let already_set_named_axes = self.already_set_named_axes(); // have named axes already been set?
+
         self.set_non_hash_info()?;
         if let Some(already_set_named_axes) = already_set_named_axes {
             self.info_mut().named_axes = already_set_named_axes; // if already set, keep these named axes
         }
+        if !prior_autoname {
+            // if disabled on init, keep disabled
+            self.info_mut().flags &= !CircuitFlags::USE_AUTONAME;
+        }
+        // we have to call maybe_auto_name after setting the flag
+        if is_initial {
+            *self.name_mut() = self.maybe_auto_name(self.name_cloned());
+        }
+        hasher.update(&[self.name().is_some() as u8]);
+        hasher.update(self.name().unwrap_or("").as_bytes());
+        hasher.update(uuid!("025e9af4-1366-4211-aa5f-7c28fc6cdf9f").as_bytes()); // delimit name with uuid
 
+        // it's important that we hash the final autoname and named axes
+        // instead of prior/already set values. (otherwise stuff like renaming
+        // to same name can change hash!)
+        hasher.update(&[self.info().use_autoname() as u8]);
         for (axis, name) in &self.info().named_axes {
             if *axis as usize >= self.info().shape.len() {
                 return Err(ConstructError::NamedAxisAboveRank {}.into());
@@ -158,20 +202,16 @@ where
         Ok(self)
     }
 
-    fn rename_impl(mut self, new_name: Option<String>) -> Self
-    where
-        Self: Sized,
-    {
+    fn rename_impl(mut self, new_name: Option<String>) -> Self {
         *self.name_mut() = new_name;
-        self.init_info_impl().unwrap() // we could avoid recomputing some stuff if we wanted
+        self.init_info_impl(false).unwrap() // we could avoid recomputing some stuff if we wanted
     }
 
     fn update_info_impl<F>(mut self, f: F) -> Result<Self>
     where
-        Self: Sized,
         F: FnOnce(&mut CachedCircuitInfo),
     {
         f(self.info_mut());
-        self.init_info_impl()
+        self.init_info_impl(false)
     }
 }
