@@ -11,7 +11,6 @@ use macro_rules_attribute::apply;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     cached_method,
-    compact_data::TinyBoundedBitvec,
     eq_by_big_hash::EqByBigHash,
     impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize, RearrangeSpec},
@@ -154,7 +153,7 @@ pub fn expand_node(
         right_align_shapes(&batch_shapes).context("batch shapes didn't match (expand)")?;
     let batch_rank = broadcasted_batch_shape.len();
 
-    let get_expanded_inputs = |batchable: TinyBoundedBitvec| -> Result<Vec<CircuitRc>, usize> {
+    let get_expanded_inputs = |batchable: &[bool]| -> Result<Vec<CircuitRc>, usize> {
         let br = batch_rank;
         assert_eq!(batchable.len(), batch_ranks.len());
         assert_eq!(inputs.len(), batch_ranks.len());
@@ -291,7 +290,7 @@ pub fn expand_node(
             // TODO: error on concat axis having symbolic size? (or maybe concat axis is fine???
             let br = batch_rank;
 
-            let inputs = get_expanded_inputs(vec![true; inputs.len()].try_into().unwrap()).unwrap();
+            let inputs = get_expanded_inputs(&vec![true; inputs.len()]).unwrap();
 
             let new_axis = node.axis + br;
             if !zip(&node.nodes, &inputs)
@@ -411,7 +410,7 @@ pub fn expand_node(
             // - [b_0, b_1, b_2, (stuff)]
             // - [b_2, (other stuff)]
             GeneralFunction::try_new(
-                get_expanded_inputs(node.input_batchability).map_err(|input_i| {
+                get_expanded_inputs(&node.input_batchability).map_err(|input_i| {
                     ExpandError::GeneralFunctionTriedToBatchNonBatchableInput {
                         input_i,
                         batched_inputs: inputs.clone(),
