@@ -23,7 +23,6 @@ pub use module::{Module, ModuleArgSpec, ModuleSpec};
 use named_axes::set_named_axes;
 use num_bigint::BigUint;
 use opaque_iterative_matcher::OpaqueIterativeMatcherVal;
-use print::PrintOptions;
 use print_html::{PrintHtmlOptions, PrintOptionsBase};
 use pyo3::{
     exceptions::{self, PyValueError},
@@ -34,7 +33,9 @@ use pyo3::{
 use rr_util::{
     cached_lambda,
     eq_by_big_hash::EqByBigHash,
-    py_types::{reduction_to_ints, use_rust_comp, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS},
+    py_types::{
+        reduction_to_ints, use_rust_comp, PyCallable, PyOpAtAxes, PyShape, Tensor, PY_CIRCUIT_ITEMS,
+    },
     pycall, pycallable, python_error_exception,
     rearrange_spec::{OpSize, PyCountsAtAxes, RInnerInts, RearrangeSpec},
     sv,
@@ -49,11 +50,14 @@ use rr_util::{
     },
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
-pub use set_symbolic_size::SetSymbolicShape;
+pub use set_symbolic_shape::SetSymbolicShape;
 use thiserror::Error;
 pub use variable_nodes::{DiscreteVar, StoredCumulantVar, Tag};
 
-use crate::{module::substitute_all_modules, print::print_circuit_stats};
+use crate::{
+    module::substitute_all_modules,
+    print::{print_circuit_stats, PrintOptions},
+};
 pub mod cached_circuit_properties;
 pub mod circuit_utils;
 pub mod computational_node;
@@ -69,7 +73,7 @@ pub mod parsing;
 pub mod print;
 pub mod print_html;
 pub mod set_of_circuits;
-pub mod set_symbolic_size;
+pub mod set_symbolic_shape;
 pub mod variable_nodes;
 
 mod circuit_node_private;
@@ -145,12 +149,6 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
 
     fn eval_tensors(&self, tensors: &[Tensor], device_dtype: &TorchDeviceDtype) -> Result<Tensor>;
 
-    /// At most how many elements will evaluating this circuit require allocating
-    /// new memory (that we are allowed to free ourselves) for? Used to improve scheduling.
-    fn intermediate_cost_bound(&self) -> usize {
-        self.info().numel_usize()
-    }
-
     fn c(self) -> Circuit;
     fn rc(self) -> CircuitRc;
 
@@ -167,9 +165,7 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
     // (if you do implement, this won't be picked up on by CircuitNodeDefer (union types etc)!)
     #[inline]
     fn compute_flags_default(&self) -> CircuitFlags {
-        let mut result: CircuitFlags = CircuitFlags::IS_EXPLICITLY_COMPUTABLE
-            | CircuitFlags::IS_CONSTANT
-            | CircuitFlags::CAN_BE_SAMPLED;
+        let mut result: CircuitFlags = CircuitFlags::all_true();
         for child in self.children() {
             result &= child.info().flags;
         }
@@ -180,11 +176,11 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
         base16::encode_lower(&self.info().hash)
     }
 
-    fn init_info(self) -> Result<Self>
+    fn initial_init_info(self) -> Result<Self>
     where
         Self: Sized,
     {
-        self.init_info_impl()
+        self.initial_init_info_impl()
     }
     fn rename(self, new_name: Option<String>) -> Self
     where
@@ -297,7 +293,20 @@ pub trait CircuitNode: CircuitNodeInit + CircuitNodeSelfOnlyHash {
             None
         }
     }
-
+    fn with_autoname_disabled(&self, autoname_disabled: bool) -> Self
+    where
+        Self: Sized + Clone,
+    {
+        self.clone()
+            .update_info(|info| {
+                if !autoname_disabled {
+                    info.flags |= CircuitFlags::USE_AUTONAME;
+                } else {
+                    info.flags &= !CircuitFlags::USE_AUTONAME;
+                }
+            })
+            .unwrap()
+    }
     fn repr(&self) -> Result<String>
     where
         Self: Clone,
@@ -612,7 +621,7 @@ pub trait CircuitNodeAutoName: CircuitNode {
     /// - functions, like GeneralFunction and Cumulant, which use the standard function notation f(x1, ... xn)
     /// - postfix, like Rearrange and index, which behave like the unary operator "!". Parenthesis are added if needed
     /// utils to add the parenthesis can be found in circuit_utils.rs
-    fn auto_name(&self, name: Option<String>) -> Option<String>;
+    fn auto_name(&self) -> Option<String>;
 
     const PRIORITY: OperatorPriority = OperatorPriority::NotOperator {};
     const CHILD_NAME_MAX_LEN: usize = 100; // TODO: fix me!
@@ -671,6 +680,14 @@ macro_rules! circuit_node_eq_ord_debug {
                 f.write_str(&self.debug_repru())
             }
         }
+
+        impl std::fmt::Display for $type_name {
+            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+                let mut printer = $crate::print::PrintOptions::default();
+                printer.bijection = false;
+                f.write_str(&printer.repr(self.crc()).unwrap())
+            }
+        }
     };
 }
 
@@ -853,9 +870,9 @@ macro_rules! define_circuit_union_impl {
 
         paste::paste! {
         impl $crate::CircuitNodeInit for $name {
-            fn init_info_impl(self) -> Result<Self> {
+            fn init_info_impl(self, is_initial: bool) -> Result<Self> {
                 [<$name:snake _map_construct>]!(self;
-                    (node, construct) => Ok(construct(node.init_info()?)))
+                    (node, construct) => Ok(construct(node.init_info_impl(is_initial)?)))
             }
 
             fn rename_impl(self, new_name: Option<String>) -> Self {
@@ -1207,10 +1224,6 @@ impl<T: CircuitNodeDefer> CircuitNode for T {
         self.as_trait_obj().eval_tensors(tensors, device_dtype)
     }
 
-    fn intermediate_cost_bound(&self) -> usize {
-        self.as_trait_obj().intermediate_cost_bound()
-    }
-
     fn c(self) -> Circuit {
         self.custom_c()
     }
@@ -1230,18 +1243,35 @@ pub struct CachedCircuitInfo {
     pub symbolic_size_constraints: SymbolicSizeConstraints,
 }
 
-#[derive(Clone, Copy, Hash, Debug, Eq, PartialEq, Default)] // note default is all false, meant to be overwritten
+#[derive(Clone, Copy, Hash, Debug, Eq, PartialEq)]
 pub struct CircuitFlags(pub u8);
 impl CircuitFlags {
-    pub const IS_CONSTANT: CircuitFlags = CircuitFlags(0b001);
-    pub const IS_EXPLICITLY_COMPUTABLE: CircuitFlags = CircuitFlags(0b010);
-    pub const CAN_BE_SAMPLED: CircuitFlags = CircuitFlags(0b100);
+    pub const IS_CONSTANT: CircuitFlags = CircuitFlags(0b00001);
+    pub const IS_EXPLICITLY_COMPUTABLE: CircuitFlags = CircuitFlags(0b00010);
+    pub const CAN_BE_SAMPLED: CircuitFlags = CircuitFlags(0b00100);
+    pub const NO_UNBOUND_SYMBOLS: CircuitFlags = CircuitFlags(0b01000);
+    pub const USE_AUTONAME: CircuitFlags = CircuitFlags(0b10000);
 
     pub const NONE: CircuitFlags = CircuitFlags(0b0);
     pub fn check(self, other: CircuitFlags) -> bool {
         (self & other).0 != 0
     }
+
+    pub fn all_true() -> Self {
+        CircuitFlags::IS_EXPLICITLY_COMPUTABLE
+            | CircuitFlags::IS_CONSTANT
+            | CircuitFlags::CAN_BE_SAMPLED
+            | CircuitFlags::NO_UNBOUND_SYMBOLS
+            | CircuitFlags::USE_AUTONAME
+    }
 }
+
+impl Default for CircuitFlags {
+    fn default() -> Self {
+        Self::all_true()
+    }
+}
+
 impl BitOr for CircuitFlags {
     type Output = CircuitFlags;
     fn bitor(self, rhs: Self) -> Self::Output {
@@ -1294,6 +1324,23 @@ impl CachedCircuitInfo {
             usize::MAX
         }
     }
+
+    pub fn naive_mem_use(&self, device_dtype: Option<TorchDeviceDtype>) -> BigUint {
+        self.numel()
+            * BigUint::from(
+                device_dtype
+                    .unwrap_or(self.device_dtype.unwrap_or_defaults())
+                    .size(),
+            )
+    }
+    // once we're scheduling everything is batch-realizeable so we don't need biguint
+    pub fn naive_mem_use_usize(&self, device_dtype: Option<TorchDeviceDtype>) -> usize {
+        self.numel_usize().saturating_mul(
+            device_dtype
+                .unwrap_or(self.device_dtype.unwrap_or_defaults())
+                .size(),
+        )
+    }
     pub fn rank(&self) -> usize {
         self.shape.len()
     }
@@ -1311,6 +1358,12 @@ impl CachedCircuitInfo {
     pub fn is_explicitly_computable(&self) -> bool {
         self.flags.check(CircuitFlags::IS_EXPLICITLY_COMPUTABLE)
     }
+    pub fn no_unbound_symbols(&self) -> bool {
+        self.flags.check(CircuitFlags::NO_UNBOUND_SYMBOLS)
+    }
+    pub fn use_autoname(&self) -> bool {
+        self.flags.check(CircuitFlags::USE_AUTONAME)
+    }
 }
 
 #[apply(python_error_exception)]
@@ -1528,6 +1581,11 @@ impl fmt::Debug for CircuitRc {
         fmt::Debug::fmt(&***self, f)
     }
 }
+impl fmt::Display for CircuitRc {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Display::fmt(&***self, f)
+    }
+}
 
 pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
     circuit.map_children_unwrap(&mut |child: CircuitRc| {
@@ -1923,8 +1981,8 @@ impl DerefMut for CircuitRc {
 }
 
 impl CircuitNodeInit for CircuitRc {
-    fn init_info_impl(self) -> Result<Self> {
-        Ok(self.c().clone().init_info()?.rc())
+    fn init_info_impl(self, is_initial: bool) -> Result<Self> {
+        Ok(self.c().clone().init_info_impl(is_initial)?.rc())
     }
 
     fn rename_impl(self, new_name: Option<String>) -> Self {
@@ -2020,6 +2078,16 @@ impl PyCircuitBase {
         self.info().can_be_sampled()
     }
 
+    #[getter]
+    fn no_unbound_symbols(&self) -> bool {
+        self.info().no_unbound_symbols()
+    }
+
+    #[getter]
+    fn use_autoname(&self) -> bool {
+        self.info().use_autoname()
+    }
+
     #[getter]
     fn name(&self) -> &str {
         self.0.name().unwrap_or("")
@@ -2030,11 +2098,6 @@ impl PyCircuitBase {
         self.0.name()
     }
 
-    #[getter]
-    fn intermediate_cost_bound(&self) -> usize {
-        self.0.intermediate_cost_bound()
-    }
-
     #[getter]
     fn symbolic_size_constraints(&self) -> SymbolicSizeConstraints {
         self.info().symbolic_size_constraints.clone()
@@ -2082,6 +2145,11 @@ impl PyCircuitBase {
         print_circuit_stats(&self.0)
     }
 
+    #[args(autoname_disabled = "true")]
+    fn with_autoname_disabled(&self, autoname_disabled: bool) -> CircuitRc {
+        self.0.clone().with_autoname_disabled(autoname_disabled)
+    }
+
     #[args(options = "Default::default()")]
     fn repr(&self, options: PrintOptions) -> Result<String> {
         options.repr((*self.0).crc())
@@ -2120,12 +2188,12 @@ impl PyCircuitBase {
         evaluate_device_dtype(self.0.clone(), device_dtype)
     }
 
-    fn map_children_enumerate(&self, f: PyObject) -> Result<CircuitRc> {
+    fn map_children_enumerate(&self, f: PyCallable) -> Result<CircuitRc> {
         self.0
             .map_children_enumerate(|i, child| pycall!(f, (i, child), anyhow))
     }
 
-    fn map_children(&self, f: PyObject) -> Result<CircuitRc> {
+    fn map_children(&self, f: PyCallable) -> Result<CircuitRc> {
         self.0.map_children(|child| pycall!(f, (child,), anyhow))
     }
 
@@ -2165,7 +2233,7 @@ impl PyCircuitBase {
         self.0.clone().add_suffix(suffix)
     }
 
-    fn visit(&self, f: PyObject) -> Result<()> {
+    fn visit(&self, f: PyCallable) -> Result<()> {
         visit_circuit_py(self.0.clone(), f)
     }
 
