@@ -661,15 +661,14 @@ impl Module {
             .substitute(&nodes, None, None)
             .with_context(|| format!("module construction failed on substitute name={:?}", name))?;
 
-        let mut out = Self {
+        let out = Self {
             nodes,
             spec,
             cached_substituted,
-            name: name.clone(),
+            name,
             info: Default::default(),
         };
-        out.name = out.auto_name(name);
-        out.init_info()
+        out.initial_init_info()
     }
 
     pub fn new_kwargs(
@@ -844,24 +843,22 @@ impl CircuitNode for Module {
 impl CircuitNodeAutoName for Module {
     const PRIORITY: OperatorPriority = OperatorPriority::Function {};
 
-    fn auto_name(&self, name: Option<String>) -> Option<String> {
-        name.or_else(|| {
-            if self.children().any(|x| x.name().is_none()) {
-                None
-            } else {
-                Some(
-                    self.spec.circuit.name_cloned().unwrap()
-                        + "("
-                        + &self
-                            .nodes
-                            .iter()
-                            .map(|node| node.name_cloned().unwrap())
-                            .collect::<Vec<String>>()
-                            .join(", ")
-                        + ")",
-                )
-            }
-        })
+    fn auto_name(&self) -> Option<String> {
+        if self.children().any(|x| x.name().is_none()) {
+            None
+        } else {
+            Some(
+                self.spec.circuit.name_cloned().unwrap()
+                    + "("
+                    + &self
+                        .nodes
+                        .iter()
+                        .map(|node| node.name_cloned().unwrap())
+                        .collect::<Vec<String>>()
+                        .join(", ")
+                    + ")",
+            )
+        }
     }
 }
 
