@@ -196,6 +196,7 @@ struct PartialCirc {
     pub children: Vec<CircuitIdent>,
     pub named_axes: Option<NamedAxes>,
     pub module_arg_specs: Vec<ModuleArgSpec>,
+    pub autoname_disabled: bool,
 }
 
 impl Parser {
@@ -249,10 +250,11 @@ impl Parser {
             num_ident_cap,
             name_ident_cap,
             name_cap,
+            autoname_disabled_cap,
             shape_cap,
             variant_cap,
             extra_cap,
-        ) = (1, 2, 3, 4, 5, 6, 7);
+        ) = (1, 2, 3, 4, 5, 6, 7, 8);
         static RE_STR: Lazy<(Regex, String)> = Lazy::new(|| {
             let name_match_capture: String = format!("'({})'", NAME_INSIDE_MATCH);
             // first match named axis name, then size. Name is optional. We handle whitespace.
@@ -260,7 +262,7 @@ impl Parser {
                 format!(r"(?:{}\s*:\s*)?{}\s*", AXIS_NAME_MATCH, *SIZE_PROD_MATCH);
             let trailing_comment_match = r"(?:#.*?)?"; // TODO: test this never matches stuff it's not supposed to...
             let full_regex = format!(
-                r"^({ws})(?:(\d+)|{nm})(?: {nm})?(?: \[((?:{sh},\s*)*(?:{sh})?)\])?(?: ([a-zA-Z]+))?(?: (.*?))?(?:{ws}){com}$",
+                r"^({ws})(?:(\d+)|{nm})(?: \s*{nm})?( \s*AD)?(?: \s*\[((?:{sh},\s*)*(?:{sh})?)\])?(?: \s*([a-zA-Z]+))?(?: \s*(.*?))?(?:{ws}){com}$",
                 ws = WHITESPACE,
                 nm = name_match_capture,
                 sh = shape_axis_match,
@@ -362,6 +364,7 @@ impl Parser {
                 .map(|(x, y)| (x.trim(), Some(y.trim())))
                 .unwrap_or((extra_and_parent_info, None));
             let extra = (!extra.is_empty()).then(|| extra.to_owned());
+            let autoname_disabled = re_captures.get(autoname_disabled_cap).is_some();
 
             let is_ref = circuit_ident
                 .as_str()
@@ -372,6 +375,7 @@ impl Parser {
                     || re_captures.get(shape_cap).is_some()
                     || re_captures.get(variant_cap).is_some()
                     || extra.is_some()
+                    || autoname_disabled
                 {
                     bail!(
                         ParseCircuitError::ReferenceCircuitNameFollowedByAdditionalInfo {
@@ -456,6 +460,7 @@ impl Parser {
                             children: vec![],
                             named_axes,
                             module_arg_specs: vec![],
+                            autoname_disabled,
                         },
                     );
                 } else if self.on_repeat_check_info_same {
@@ -486,6 +491,8 @@ impl Parser {
                     fail_check!(shape, Some);
                     fail_check!(variant);
                     fail_check!(extra);
+                    let autoname_disabled = autoname_disabled.then_some(true);
+                    fail_check!(autoname_disabled);
                     if any_fail {
                         bail!(ParseCircuitError::OnCircuitRepeatInfoIsNotSame {
                             repeat_ident: format!("{:?}", circuit_ident),
@@ -992,6 +999,12 @@ impl Parser {
         } else {
             result
         };
+        let result = if ps.autoname_disabled {
+            result.with_autoname_disabled(true)
+        } else {
+            result
+        };
+
         Ok(result)
     }
 }
