@@ -14,7 +14,7 @@ use once_cell::unsync::Lazy;
 use pyo3::{once_cell::GILLazy, prelude::*};
 use rr_util::{
     fn_struct, py_types::MaybeNotSet, python_println, symbolic_size::SymbolicSizeProduct,
-    util::ALPHABET,
+    tensor_db::write_tensor_to_dir_tree, util::ALPHABET,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
@@ -135,8 +135,6 @@ pub struct PrintOptions {
     #[pyo3(get, set)]
     pub arrows: bool,
     #[pyo3(get, set)]
-    pub print_size_threshold: Option<usize>,
-    #[pyo3(get, set)]
     pub force_use_serial_numbers: bool,
     #[pyo3(get, set)]
     pub number_leaves: bool,
@@ -156,6 +154,8 @@ pub struct PrintOptions {
     pub tensor_index_literal: bool,
     #[pyo3(get, set)]
     pub show_all_named_axes: bool,
+    #[pyo3(get, set)]
+    pub tensor_save_dir: Option<String>,
 }
 
 impl Default for PrintOptions {
@@ -167,7 +167,6 @@ impl Default for PrintOptions {
             traversal: None,
             leaves_on_top: false,
             arrows: false,
-            print_size_threshold: None,
             force_use_serial_numbers: false,
             number_leaves: false,
             colorer: None,
@@ -178,6 +177,7 @@ impl Default for PrintOptions {
             only_child_below: false,
             tensor_index_literal: false,
             show_all_named_axes: false,
+            tensor_save_dir: None,
         }
         .validate_ret()
         .unwrap()
@@ -316,7 +316,6 @@ impl PrintOptions {
 }
 
 const DEFAULT_END_DEPTH: usize = 2;
-const DEFAULT_PRINT_SIZE_THRESHOLD: usize = 400_000_000;
 
 #[derive(Debug, Clone)]
 pub struct TerseBool(pub bool);
@@ -387,7 +386,6 @@ impl PrintOptions {
         traversal = "None",
         leaves_on_top = "PrintOptions::default().leaves_on_top",
         arrows = "PrintOptions::default().arrows",
-        print_size_threshold = "None",
         force_use_serial_numbers = "PrintOptions::default().force_use_serial_numbers",
         number_leaves = "PrintOptions::default().number_leaves",
         colorer = "None",
@@ -397,7 +395,8 @@ impl PrintOptions {
         seen_children_same_line = "PrintOptions::default().seen_children_same_line",
         only_child_below = "PrintOptions::default().only_child_below",
         tensor_index_literal = "PrintOptions::default().tensor_index_literal",
-        show_all_named_axes = "PrintOptions::default().show_all_named_axes"
+        show_all_named_axes = "PrintOptions::default().show_all_named_axes",
+        tensor_save_dir = "None"
     )]
     pub fn new(
         bijection: bool,
@@ -407,7 +406,6 @@ impl PrintOptions {
         traversal: Option<OpaqueIterativeMatcherVal>,
         leaves_on_top: bool,
         arrows: bool,
-        print_size_threshold: Option<usize>,
         force_use_serial_numbers: bool,
         number_leaves: bool,
         colorer: Option<CircuitToColorCode>,
@@ -418,6 +416,7 @@ impl PrintOptions {
         only_child_below: bool,
         tensor_index_literal: bool,
         show_all_named_axes: bool,
+        tensor_save_dir: Option<String>,
     ) -> Result<Self> {
         let result = Self {
             bijection,
@@ -435,7 +434,6 @@ impl PrintOptions {
             traversal,
             leaves_on_top,
             arrows,
-            print_size_threshold,
             force_use_serial_numbers,
             number_leaves,
             colorer,
@@ -446,6 +444,7 @@ impl PrintOptions {
             only_child_below,
             tensor_index_literal,
             show_all_named_axes,
+            tensor_save_dir,
         };
         result.validate()?;
         Ok(result)
@@ -458,7 +457,6 @@ impl PrintOptions {
         traversal = "Default::default()",
         leaves_on_top = "Default::default()",
         arrows = "Default::default()",
-        print_size_threshold = "Default::default()",
         force_use_serial_numbers = "Default::default()",
         number_leaves = "Default::default()",
         colorer = "Default::default()",
@@ -468,7 +466,8 @@ impl PrintOptions {
         seen_children_same_line = "Default::default()",
         only_child_below = "Default::default()",
         tensor_index_literal = "Default::default()",
-        show_all_named_axes = "Default::default()"
+        show_all_named_axes = "Default::default()",
+        tensor_save_dir = "Default::default()"
     )]
     pub fn evolve(
         &self,
@@ -478,7 +477,6 @@ impl PrintOptions {
         traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
         leaves_on_top: MaybeNotSet<bool>,
         arrows: MaybeNotSet<bool>,
-        print_size_threshold: MaybeNotSet<Option<usize>>,
         force_use_serial_numbers: MaybeNotSet<bool>,
         number_leaves: MaybeNotSet<bool>,
         colorer: MaybeNotSet<Option<CircuitToColorCode>>,
@@ -489,6 +487,7 @@ impl PrintOptions {
         only_child_below: MaybeNotSet<bool>,
         tensor_index_literal: MaybeNotSet<bool>,
         show_all_named_axes: MaybeNotSet<bool>,
+        tensor_save_dir: MaybeNotSet<Option<String>>,
     ) -> Result<Self> {
         let cloned = self.clone();
         Self::new(
@@ -501,9 +500,6 @@ impl PrintOptions {
             traversal.0.unwrap_or(cloned.traversal),
             leaves_on_top.0.unwrap_or(cloned.leaves_on_top),
             arrows.0.unwrap_or(cloned.arrows),
-            print_size_threshold
-                .0
-                .unwrap_or(cloned.print_size_threshold),
             force_use_serial_numbers
                 .0
                 .unwrap_or(cloned.force_use_serial_numbers),
@@ -520,6 +516,7 @@ impl PrintOptions {
                 .0
                 .unwrap_or(cloned.tensor_index_literal),
             show_all_named_axes.0.unwrap_or(cloned.show_all_named_axes),
+            tensor_save_dir.0.unwrap_or(cloned.tensor_save_dir),
         )
     }
 
@@ -536,7 +533,6 @@ impl PrintOptions {
     #[staticmethod]
     pub fn compiler_default() -> PrintOptions {
         Self {
-            print_size_threshold: Some(DEFAULT_PRINT_SIZE_THRESHOLD),
             force_use_serial_numbers: true,
             ..Default::default()
         }
@@ -601,12 +597,8 @@ impl PrintOptions {
         if self.bijection && self.tensor_index_literal {
             bail!("bijection doesn't support index tensor literal")
         }
-        if self.seen_children_same_line
-            && (self.comment_arg_names
-                || !self.commenters.is_empty()
-                || self.print_size_threshold.is_some())
-        {
-            bail!("children same line + commenting (arg names, commenter, print_size_threshold) not supported")
+        if self.seen_children_same_line && (self.comment_arg_names || !self.commenters.is_empty()) {
+            bail!("children same line + commenting (arg names, commenter) not supported")
         }
         if self.bijection && self.is_print_as_tree() {
             bail!("bijection print can't print as tree (caused by number_leaves)")
@@ -632,6 +624,10 @@ impl PrintOptions {
     pub fn repr_line_info(&self, circ: CircuitRc) -> Result<String> {
         let mut result = "".to_owned();
 
+        if self.bijection && !circ.info().use_autoname() {
+            result.push_str(" AD");
+        }
+
         if !self.shape_only_when_necessary
             || matches!(
                 &**circ,
@@ -718,7 +714,11 @@ impl PrintOptions {
                         ac.save_rrfs(true)?;
                     }
                     if self.bijection {
-                        ac.save_rrfs(false)?;
+                        if let Some(save_dir) = &self.tensor_save_dir {
+                            write_tensor_to_dir_tree(save_dir, ac.value.clone(), false)?;
+                        } else {
+                            ac.save_rrfs(false)?;
+                        }
                         ac.tensor_hash_base16()[..24].to_owned()
                     } else {
                         "".to_owned()
@@ -883,11 +883,6 @@ impl PrintOptions {
                 *runnning_leaf_number += 1;
             }
 
-            // TODO: move to commenter
-            if let Some(threshold) = selfy.print_size_threshold && circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
-                node_comments.push(color(&oom_fmt(circ.info().numel()), clicolor!(Red)));
-            }
-
             write_comment(result, node_comments, &selfy.commenters, circ.crc())?;
             result.push('\n');
 
@@ -1096,6 +1091,30 @@ impl PrintOptions {
             Ok(map.get(&c.info().hash).cloned().into())
         }))
     }
+
+    #[staticmethod]
+    #[args(threshold = "400_000_000")]
+    pub fn size_threshold_commenter(threshold: usize) -> CircuitCommenter {
+        CircuitCommenter::new_dyn(Box::new(move |circ| {
+            if circ.info().numel() > BigUint::from(threshold) && !circ.is_array() {
+                Ok(color(&oom_fmt(circ.info().numel()), clicolor!(Red)))
+            } else {
+                Ok("".to_string())
+            }
+        }))
+    }
+
+    #[staticmethod]
+    #[args("*", only_arrays = "false")]
+    pub fn dtype_commenter(only_arrays: bool) -> CircuitCommenter {
+        CircuitCommenter::new_dyn(Box::new(move |circ| {
+            if let Some(dtype) = &circ.info().device_dtype.dtype && !(only_arrays && !circ.is_array()) {
+                Ok(dtype.into())
+            } else {
+                Ok("".to_string())
+            }
+        }))
+    }
 }
 
 fn init_print_options() -> PrintOptions {
