@@ -8,7 +8,7 @@ use anyhow::{bail, Context, Error, Result};
 use bit_iter::BitIter;
 use circuit_base::{
     computational_node::EinsumArgs, prelude::*, Add, CircuitType, Concat, DiscreteVar, Einsum,
-    GeneralFunction, Index, Rearrange, Scalar, Scatter, Symbol, Tag,
+    GeneralFunction, Index, Module, Rearrange, Scalar, Scatter, Symbol, Tag,
 };
 use itertools::Itertools;
 use macro_rules_attribute::apply;
@@ -331,7 +331,7 @@ pub fn einsum_nest_optimize(einsum: &Einsum, context: &mut OptimizationContext)
     let path = optimize_einsum_spec_cached(
         einsum.get_spec(),
         None,
-        Some(context.cache.max_single_tensor_numel),
+        Some(context.settings.max_single_tensor_memory / einsum.info().device_dtype.size()),
         None,
         Some(filter_to_idx(&einsum.args, &|(x, _)| {
             matches!(&***x, Circuit::Array(_))
@@ -499,7 +499,13 @@ pub fn index_fuse(index: &Index) -> Option<Index> {
         Some(Index::new(
             inner.node.clone(),
             compose(&index.index, &inner.index)?,
-            index.name_cloned(),
+            index.name_cloned().map(|n| {
+                if n.ends_with(" idx idx") {
+                    n.strip_suffix(" idx").unwrap().to_owned()
+                } else {
+                    n
+                }
+            }),
         ))
     })
 }
@@ -902,14 +908,14 @@ pub fn distribute_once_raw(
     operand_idx: usize,
     do_broadcasts: bool,
     suffix: Option<String>,
-    mut call_on_sub: impl FnMut(Einsum) -> Result<CircuitRc>,
+    mut call_on_sub: impl FnMut(usize, Einsum) -> Result<CircuitRc>,
 ) -> Result<Add> {
-    let mut on_sub = |node: CircuitRc, einsum: Einsum| -> Result<CircuitRc> {
+    let mut on_sub = |child_idx, node: CircuitRc, einsum: Einsum| -> Result<CircuitRc> {
         let new_name = suffix
             .as_ref()
             .map(|s| node.name().map(|n_s| format!("{n_s}{s}")))
             .flatten();
-        call_on_sub(einsum.rename(new_name))
+        call_on_sub(child_idx, einsum.rename(new_name))
     };
     match einsum.args[operand_idx].0.as_add() {
         Some(add_node) => {
@@ -929,10 +935,12 @@ pub fn distribute_once_raw(
             let summands = add
                 .nodes
                 .iter()
-                .map(|node| {
+                .enumerate()
+                .map(|(i, node)| {
                     let mut new_args = einsum.args.clone();
                     new_args[operand_idx].0 = node.clone();
                     on_sub(
+                        i,
                         node.clone(),
                         Einsum::new(new_args, einsum.out_axes.clone(), None),
                     )
@@ -960,11 +968,17 @@ pub fn distribute_once(
             operand_idx,
         })
     }
-    distribute_once_raw(einsum, operand_idx, do_broadcasts, suffix, |c| Ok(c.rc()))
+    distribute_once_raw(
+        einsum,
+        operand_idx,
+        do_broadcasts,
+        suffix,
+        |_, c| Ok(c.rc()),
+    )
 }
 
 pub fn distribute_once_op(einsum: &Einsum, operand_idx: usize, do_broadcasts: bool) -> Option<Add> {
-    distribute_once_raw(einsum, operand_idx, do_broadcasts, None, |c| Ok(c.rc())).ok()
+    distribute_once_raw(einsum, operand_idx, do_broadcasts, None, |_, c| Ok(c.rc())).ok()
 }
 
 /// distribute all Adds that are direct children of this einsum
@@ -1235,42 +1249,25 @@ pub fn axis_index_split_sections_for_concat(
     Ok(out)
 }
 
-// #[pyfunction]
-// pub fn push_down_index_small(node:&Index)->Option<CircuitRc>{
-//         // if you don't specify which_axes, choose all axes that reduce dimension
-//     let which_axes = which_axes.unwrap_or_else(|| {
-//         zip(node.index.0, node.info().shape)
-//             .enumerate()
-//             .filter_map(|(i, (idx, l))| {
-//                 if let TensorAxisIndex::Tensor(tensor) = idx {
-//                     if tensor.shape()[0] >= l {
-//                         None
-//                     } else {
-//                         Some(i)
-//                     }
-//                 } else {
-//                     Some(i)
-//                 }
-//             }).collect()
-//     });
-// }
-//
 pub fn push_down_index_raw(
     node: &Index,
     allow_partial_pushdown: bool,
-    mut call_on_sub: impl FnMut(Index) -> Result<CircuitRc>,
+    mut call_on_sub: impl FnMut(usize, Index) -> Result<CircuitRc>,
     suffix: Option<String>,
 ) -> Result<CircuitRc> {
-    let mut on_sub =
-        |node: CircuitRc, index, force_name: Option<Option<String>>| -> Result<CircuitRc> {
-            let new_name = force_name.unwrap_or_else(|| {
-                suffix
-                    .as_ref()
-                    .map(|s| node.name().map(|n_s| format!("{n_s}{s}")))
-                    .flatten()
-            });
-            call_on_sub(Index::new(node, index, new_name))
-        };
+    let mut on_sub = |child_idx,
+                      node: CircuitRc,
+                      index,
+                      force_name: Option<Option<String>>|
+     -> Result<CircuitRc> {
+        let new_name = force_name.unwrap_or_else(|| {
+            suffix
+                .as_ref()
+                .map(|s| node.name().map(|n_s| format!("{n_s}{s}")))
+                .flatten()
+        });
+        call_on_sub(child_idx, Index::new(node, index, new_name))
+    };
     let out_name = node.name_cloned();
     let partial_name = || {
         suffix
@@ -1282,6 +1279,37 @@ pub fn push_down_index_raw(
             })
             .flatten()
     };
+
+    let handle_non_batchable_remaining = |rank_to_pass, new_circ: CircuitRc| {
+        let passed_rank_now = rank_to_pass
+            - (0..rank_to_pass)
+                .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
+                .count();
+        let top_axis_indices: Vec<_> = node.index.0[rank_to_pass..].iter().cloned().collect();
+        if top_axis_indices
+            .iter()
+            .zip(&new_circ.shape()[passed_rank_now..])
+            .all(|(x, s)| x.is_identity(*s))
+        {
+            Ok(new_circ)
+        } else if allow_partial_pushdown {
+            let index_top = TensorIndex(
+                vec![TensorAxisIndex::IDENT; passed_rank_now]
+                    .iter()
+                    .cloned()
+                    .chain(top_axis_indices)
+                    .collect(),
+            );
+            Ok(Index::nrc(
+                new_circ.rename(partial_name()),
+                index_top,
+                out_name.clone(),
+            ))
+        } else {
+            Err((passed_rank_now, top_axis_indices))
+        }
+    };
+
     // TODO: maybe handle symbolic axes? Possibilities:
     // Check we don't index into sum of symbolic axes
     // Handle rearrange case.
@@ -1290,7 +1318,8 @@ pub fn push_down_index_raw(
             let new_operands = inner
                 .nodes_and_rank_differences()
                 .iter()
-                .map(|(operand, rank_difference)| {
+                .enumerate()
+                .map(|(i, (operand, rank_difference))| {
                     let index_here = (0..operand.info().rank())
                         .map(|i| {
                             let idx_here = node.index.0[i + rank_difference].clone();
@@ -1304,7 +1333,7 @@ pub fn push_down_index_raw(
                             }
                         })
                         .collect();
-                    on_sub(operand.clone(), TensorIndex(index_here), None)
+                    on_sub(i, operand.clone(), TensorIndex(index_here), None)
                 })
                 .collect::<Result<_>>()?;
             Add::nrc(new_operands, out_name)
@@ -1364,18 +1393,19 @@ pub fn push_down_index_raw(
                         // should only be once match
                         if let Some(z) = ax {
                             index.0[inner.axis] = z.clone();
-                            return on_sub(inner.nodes[i].clone(), index, Some(out_name));
+                            return on_sub(i, inner.nodes[i].clone(), index, Some(out_name));
                         }
                     }
                     unreachable!()
                 } else {
                     Concat::nrc(
                         zip(inner.nodes.iter(), concat_indices)
-                            .filter_map(|(child, idx)| {
+                            .enumerate()
+                            .filter_map(|(i, (child, idx))| {
                                 idx.map(|idx| {
                                     let mut index = index_non_axis.clone();
                                     index.0[inner.axis] = idx;
-                                    on_sub(child.clone(), index, None)
+                                    on_sub(i, child.clone(), index, None)
                                 })
                             })
                             .collect::<Result<_>>()?,
@@ -1388,7 +1418,10 @@ pub fn push_down_index_raw(
                     inner
                         .nodes
                         .iter()
-                        .map(|operand| on_sub(operand.clone(), index_non_axis.clone(), None))
+                        .enumerate()
+                        .map(|(i, operand)| {
+                            on_sub(i, operand.clone(), index_non_axis.clone(), None)
+                        })
                         .collect::<Result<_>>()?,
                     new_axis,
                     partial_name(),
@@ -1414,8 +1447,9 @@ pub fn push_down_index_raw(
             } else {
                 let rank_to_pass =
                     inner.info().rank() - inner.num_non_batchable_output_dims as usize;
-                let new_operands = zip(&inner.nodes, inner.input_batchability)
-                    .map(|(child, batchable)| {
+                let new_operands = zip(&inner.nodes, &inner.input_batchability)
+                    .enumerate()
+                    .map(|(i, (child, &batchable))| {
                         if batchable {
                             let ident_rank = child.info().rank() - rank_to_pass;
                             let index_passed = TensorIndex(
@@ -1425,7 +1459,7 @@ pub fn push_down_index_raw(
                                     .chain(vec![TensorAxisIndex::IDENT; ident_rank])
                                     .collect(),
                             );
-                            on_sub(child.clone(), index_passed, None)
+                            on_sub(i, child.clone(), index_passed, None)
                         } else {
                             Ok(child.clone())
                         }
@@ -1434,36 +1468,17 @@ pub fn push_down_index_raw(
                 let new_gf =
                     GeneralFunction::nrc(new_operands, inner.spec.clone(), out_name.clone());
 
-                let passed_rank_now = rank_to_pass
-                    - (0..rank_to_pass)
-                        .filter(|i| matches!(node.index.0[*i], TensorAxisIndex::Single(_)))
-                        .count();
-                let top_axis_indices: Vec<_> =
-                    node.index.0[rank_to_pass..].iter().cloned().collect();
-                if top_axis_indices
-                    .iter()
-                    .zip(&new_gf.shape()[passed_rank_now..])
-                    .all(|(x, s)| x.is_identity(*s))
-                {
-                    new_gf
-                } else if allow_partial_pushdown {
-                    let index_top = TensorIndex(
-                        vec![TensorAxisIndex::IDENT; passed_rank_now]
-                            .iter()
-                            .cloned()
-                            .chain(top_axis_indices)
-                            .collect(),
-                    );
-                    Index::nrc(new_gf.rename(partial_name()), index_top, out_name)
-                } else {
-                    bail!(PushDownIndexError::GeneralFunctionSomeAxesNotPossible {
-                        index_node: node.clone(),
-                        is_batchable: inner.is_batchable(),
-                        num_non_batch_out: inner.num_non_batchable_output_dims,
-                        inner_rank: inner.rank(),
-                        top_axis_indices
-                    })
-                }
+                handle_non_batchable_remaining(rank_to_pass, new_gf).map_err(
+                    |(_, top_axis_indices)| {
+                        PushDownIndexError::GeneralFunctionSomeAxesNotPossible {
+                            index_node: node.clone(),
+                            is_batchable: inner.is_batchable(),
+                            num_non_batch_out: inner.num_non_batchable_output_dims,
+                            inner_rank: inner.rank(),
+                            top_axis_indices,
+                        }
+                    },
+                )?
             }
         }
         Circuit::Einsum(inner) => {
@@ -1498,7 +1513,8 @@ pub fn push_down_index_raw(
             let new_operands: Vec<(CircuitRc, EinsumAxes)> = inner
                 .args
                 .iter()
-                .map(|(node, ints)| {
+                .enumerate()
+                .map(|(i, (node, ints))| {
                     let index_here = TensorIndex(
                         ints.iter()
                             .map(|i| {
@@ -1514,7 +1530,7 @@ pub fn push_down_index_raw(
                         any_operand_index_nontrivial = true;
                     }
                     Ok((
-                        on_sub(node.clone(), index_here, None)?,
+                        on_sub(i, node.clone(), index_here, None)?,
                         ints_filter_index(ints), // todo make function work with sv
                     ))
                 })
@@ -1581,7 +1597,7 @@ pub fn push_down_index_raw(
                     spec: (*inner.spec).clone()
                 })
             };
-            let new_index = on_sub(inner.node.clone(), index, None)?;
+            let new_index = on_sub(0, inner.node.clone(), index, None)?;
             Rearrange::nrc_elim_identity(new_index, spec, out_name)
         }
         Circuit::Scatter(inner) => {
@@ -1612,7 +1628,7 @@ pub fn push_down_index_raw(
                     index_node: node.clone()
                 });
             }
-            let inner_node = on_sub(inner.node.clone(), TensorIndex(passed_index), None)?;
+            let inner_node = on_sub(0, inner.node.clone(), TensorIndex(passed_index), None)?;
 
             let new_scatter = Scatter::nrc(
                 inner_node,
@@ -1635,12 +1651,13 @@ pub fn push_down_index_raw(
             }
         }
         Circuit::Tag(tag) => Tag::nrc(
-            on_sub(tag.node.clone(), node.index.clone(), None)?,
+            on_sub(0, tag.node.clone(), node.index.clone(), None)?,
             tag.uuid,
             out_name,
         ),
         Circuit::DiscreteVar(var) => DiscreteVar::nrc(
             on_sub(
+                0,
                 var.values.clone(),
                 TensorIndex(
                     iter::once(TensorAxisIndex::IDENT) // skip sample axis
@@ -1653,6 +1670,41 @@ pub fn push_down_index_raw(
             out_name,
         ),
         Circuit::Scalar(_) => index_merge_scalar(node).unwrap(),
+        Circuit::Module(m) => {
+            // doesn't push anything into spec,
+            let batch_shape = m.aligned_batch_shape();
+            let rank_to_pass = batch_shape.len();
+
+            let new_nodes = m
+                .nodes
+                .iter()
+                .zip(&m.spec.arg_specs)
+                .enumerate()
+                .map(|(i, (input_node, arg_spec))| {
+                    let this_num_batch = input_node.rank() - arg_spec.symbol.rank();
+                    assert!(this_num_batch <= rank_to_pass);
+                    let start = rank_to_pass - this_num_batch;
+                    let index_passed = TensorIndex(
+                        node.index.0[start..rank_to_pass]
+                            .iter()
+                            .cloned()
+                            .chain(vec![TensorAxisIndex::IDENT; arg_spec.symbol.rank()])
+                            .collect(),
+                    );
+                    on_sub(i + 1, input_node.clone(), index_passed, None)
+                })
+                .collect::<Result<_>>()?;
+            let new_mod = Module::nrc(new_nodes, m.spec.clone(), out_name.clone());
+            handle_non_batchable_remaining(rank_to_pass, new_mod).map_err(
+                |(_, top_axis_indices)| PushDownIndexError::ModuleSomeAxesNotPossible {
+                    index_node: node.clone(),
+                    num_non_batch_out: m.rank() - rank_to_pass,
+                    inner_rank: m.rank(),
+                    top_axis_indices,
+                },
+            )?
+        }
+
         // Circuit::Cumulant(cum) => Cumulant::nrc(
         //     cum.dims_for_self().map(...)
         //     out_name,
@@ -1679,13 +1731,13 @@ pub fn push_down_index_once(
     push_down_index_raw(
         &node,
         allow_partial_pushdown.unwrap_or(false),
-        |x| Ok(x.rc()),
+        |_, x| Ok(x.rc()),
         suffix,
     )
 }
 
 pub fn push_down_index_op(node: &Index) -> Option<CircuitRc> {
-    push_down_index_raw(node, true, |x| Ok(x.rc()), None).ok()
+    push_down_index_raw(node, true, |_, x| Ok(x.rc()), None).ok()
 }
 
 #[pyfunction]
@@ -1970,6 +2022,11 @@ pub fn einsum_permute_to_rearrange(einsum: &Einsum) -> Option<Rearrange> {
     None
 }
 
+const MODULE_AXIS_NO_MSG: &str = concat!(
+    "Some axes would be pushed into spec circuit, so aren't possible.\n",
+    "You could substitute the module if you want to push into the spec circuit."
+);
+
 #[apply(python_error_exception)]
 #[base_error_name(PushDownIndex)]
 #[base_exception(PyValueError)]
@@ -1996,6 +2053,13 @@ pub enum PushDownIndexError {
         inner_rank: usize,
         top_axis_indices: Vec<TensorAxisIndex>,
     },
+    #[error("{MODULE_AXIS_NO_MSG}\nindex_node={index_node:?} num_non_batch_out={num_non_batch_out} inner_rank={inner_rank} top_axis_indices={top_axis_indices:?} ({e_name})")]
+    ModuleSomeAxesNotPossible {
+        index_node: Index,
+        num_non_batch_out: usize,
+        inner_rank: usize,
+        top_axis_indices: Vec<TensorAxisIndex>,
+    },
     // TODO: improve errors as needed
     #[error("index_node={index_node:?} ({e_name})")]
     EinsumNoop { index_node: Index },
