@@ -193,11 +193,12 @@ pub fn batch_einsum_py(einsum: Einsum, settings: OptimizationSettings) -> Result
 }
 
 pub fn batch_einsum(einsum: &Einsum, context: &mut OptimizationContext) -> Result<CircuitRc> {
-    let max_single_mem = context.cache.max_single_tensor_numel;
+    let max_single_mem =
+        context.settings.max_single_tensor_memory / einsum.info().device_dtype.size();
     assert!(einsum.info().numel_usize() <= max_single_mem);
 
     let mut no_memory_limit_context: OptimizationContext = Default::default();
-    no_memory_limit_context.cache.max_single_tensor_numel = usize::MAX;
+    no_memory_limit_context.settings.max_single_tensor_memory = usize::MAX;
     // use named axes to record the ints used in intermediate einsums
     let einsum_args_named_axes: Vec<(CircuitRc, EinsumAxes)> = einsum
         .args
