@@ -60,8 +60,6 @@ pub struct OptimizationSettings {
     #[pyo3(get, set)]
     pub scheduling_timeout: usize,
     #[pyo3(get, set)]
-    pub scheduling_simplify_lossy: bool,
-    #[pyo3(get, set)]
     pub adjust_numerical_scale: bool,
     #[pyo3(get, set)]
     pub numerical_scale_min: f64,
@@ -91,7 +89,6 @@ impl Default for OptimizationSettings {
             scheduling_naive: false,
             scheduling_timeout: 5_000,
             scheduling_simplify: true,
-            scheduling_simplify_lossy: false,
             adjust_numerical_scale: false,
             numerical_scale_min: 1e-8,
             numerical_scale_max: 1e8,
@@ -117,7 +114,6 @@ impl OptimizationSettings {
         scheduling_naive = "OptimizationSettings::default().scheduling_naive",
         scheduling_timeout = "OptimizationSettings::default().scheduling_timeout",
         scheduling_simplify = "OptimizationSettings::default().scheduling_simplify",
-        scheduling_simplify_lossy = "OptimizationSettings::default().scheduling_simplify_lossy",
         adjust_numerical_scale = "OptimizationSettings::default().adjust_numerical_scale",
         numerical_scale_min = "OptimizationSettings::default().numerical_scale_min",
         numerical_scale_max = "OptimizationSettings::default().numerical_scale_max",
@@ -137,7 +133,6 @@ impl OptimizationSettings {
         scheduling_naive: bool,
         scheduling_timeout: usize,
         scheduling_simplify: bool,
-        scheduling_simplify_lossy: bool,
         adjust_numerical_scale: bool,
         numerical_scale_min: f64,
         numerical_scale_max: f64,
@@ -157,7 +152,6 @@ impl OptimizationSettings {
             scheduling_naive,
             scheduling_timeout,
             scheduling_simplify,
-            scheduling_simplify_lossy,
             adjust_numerical_scale,
             numerical_scale_min,
             numerical_scale_max,
@@ -184,9 +178,6 @@ pub struct OptimizationCache {
     pub stripped_names: HashMap<HashBytes, Option<CircuitRc>>,
     pub times_distributed: usize,
     pub simplification_log: Vec<&'static str>,
-    pub max_tensor_elements: usize,
-    pub max_single_tensor_numel: usize,
-    pub fallback_total_numel: usize,
     pub slow_einsum_log: Vec<EinsumSpec>,
 }
 
@@ -214,10 +205,6 @@ impl OptimizationContext {
             settings,
         };
         let dtype = get_compatible_dtype(&circuit);
-        result.cache.max_tensor_elements = settings.max_memory / dtype.size();
-        result.cache.max_single_tensor_numel = settings.max_single_tensor_memory / dtype.size();
-        result.cache.fallback_total_numel =
-            settings.max_memory_fallback.unwrap_or(settings.max_memory) / dtype.size();
         result
     }
 
@@ -247,7 +234,7 @@ pub fn optimize_circuit(
     context: &mut OptimizationContext,
 ) -> Result<CircuitRc> {
     let print_timings = context.settings.verbose >= 2;
-    let circuit = timed!(strip_names(circuit, context), 10, print_timings);
+    let circuit = timed!(strip_names(circuit, context, true), 10, print_timings);
     let circuit = remove_autotags(circuit);
     let circuit = conform_all_modules(circuit);
     if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
@@ -279,6 +266,7 @@ pub fn optimize_circuit(
         apply_fn_until_same(&circuit, |x| deep_module_remove_unused_inputs(
             compiler_simp(x.clone(), context),
             &mut context.cache.module_removed_unused,
+            false,
             true
         )),
         10,
