@@ -3,7 +3,7 @@ use std::fmt::{self, Display};
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
     cached_circuit_properties::max_non_leaf_size,
-    circuit_utils::{hash_to_node_non_free, toposort_circuit},
+    circuit_utils::{hash_to_node_non_free, is_definitely_view_on_child, toposort_circuit},
     get_compatible_dtype,
     module::{any_children_with_symbolic_sizes, substitute_all_modules},
     prelude::*,
@@ -342,8 +342,12 @@ pub enum SchedulingOOMError {
         node_memories: Vec<usize>,
         string: String,
     },
-    #[error("Single element doesn't fit {numel} ({e_name})")]
-    Single { numel: BigUint },
+    #[error("Single element doesn't fit {formatted} ({e_name})")]
+    Single { numel: BigUint, formatted: String },
+    #[error(
+        "Simplification can't find anything within bounds, at {origin} graph {cyto} ({e_name})"
+    )]
+    Simp { cyto: String, origin: String },
 }
 
 #[apply(python_error_exception)]
@@ -679,19 +683,25 @@ pub fn order_to_schedule(
 
 pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
     let mut result: Dag = Default::default();
+    let mut to_merge_with_child: Vec<u32> = vec![];
     // append the circuit to the dag if it's not already there, and return its index
-    let number_node = |c: CircuitRc, result: &mut Dag| -> u32 {
+    let number_node = |c: CircuitRc, result: &mut Dag, to_merge_with_child: &mut Vec<u32>| -> u32 {
         if let Some(idx) = result.hash_to_node.get(&c.info().hash) {
             return *idx as u32;
         }
-        result.node_costs.push(c.intermediate_cost_bound());
+        result.node_costs.push(c.info().naive_mem_use_usize(None));
         result.node_hashes.push(c.info().hash);
         result
             .hash_to_node
             .insert(c.info().hash, result.node_hashes.len() - 1);
         result.children.push(sv![]);
         result.parents.push(sv![]);
-        (result.node_hashes.len() - 1) as u32
+        let output = (result.node_hashes.len() - 1) as u32;
+
+        if is_definitely_view_on_child(c.clone()) {
+            to_merge_with_child.push(output)
+        }
+        output
     };
     for circuit in circuits {
         // non free so we avoid recuring into module spec.circuit
@@ -700,8 +710,9 @@ pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
             |c: CircuitRc| {
                 // Arrays are never added to the dag to begin with because they're always 0 cost
 
-                if !Leaf::matches(&c) {
-                    let my_number: u32 = number_node(c.clone(), &mut result);
+                if !c.is_leaf() {
+                    let my_number: u32 =
+                        number_node(c.clone(), &mut result, &mut to_merge_with_child);
 
                     let children_to_consider: Vec<CircuitRc> = c
                         .non_free_children()
@@ -709,11 +720,14 @@ pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
                         .collect();
                     result.children[my_number as usize] = children_to_consider
                         .iter()
-                        .map(|child| number_node(child.clone(), &mut result))
+                        .map(|child| {
+                            number_node(child.clone(), &mut result, &mut to_merge_with_child)
+                        })
                         .unique()
                         .collect();
                     for child in children_to_consider {
-                        let new_num = number_node(child.clone(), &mut result);
+                        let new_num =
+                            number_node(child.clone(), &mut result, &mut to_merge_with_child);
                         if !result.parents[new_num as usize]
                             .iter()
                             .contains(&(my_number as u32))
@@ -732,6 +746,12 @@ pub fn circuits_to_dag(circuits: &[CircuitRc]) -> Dag {
     result.node_to_orig = (0..result.node_costs.len() as u32)
         .map(|x| (x, sv![x].clone()))
         .collect();
+    for tmp in to_merge_with_child {
+        assert!(result.children[tmp as usize].len() < 2);
+        if result.children[tmp as usize].len() == 1 {
+            result.merge_larger(result.children[tmp as usize][0], tmp, false)
+        }
+    }
     result
 }
 
@@ -741,8 +761,12 @@ pub fn circuit_to_schedule(
 ) -> Result<Schedule> {
     // we could avoid this if we wanted, but a bit annoying
     // (requires handle rearrange due to unused arg + empty instruction case!)
-    let circuit =
-        deep_module_remove_unused_inputs(circuit, &mut context.cache.module_removed_unused, true);
+    let circuit = deep_module_remove_unused_inputs(
+        circuit,
+        &mut context.cache.module_removed_unused,
+        true,
+        true,
+    );
     if is_python_running() && any_children_with_symbolic_sizes(circuit.clone()) {
         Python::with_gil(|py| {
             pyo3::PyErr::warn(
@@ -759,9 +783,11 @@ pub fn circuit_to_schedule(
     }
 
     let mut dag = circuits_to_dag(&[circuit.clone()]);
-    if max_non_leaf_size(circuit.clone()) > BigUint::from(context.cache.fallback_total_numel) {
+    if max_non_leaf_size(circuit.clone()) > BigUint::from(context.settings.max_single_tensor_memory)
+    {
         bail!(SchedulingOOMError::Single {
             numel: max_non_leaf_size(circuit.clone()),
+            formatted: oom_fmt(max_non_leaf_size(circuit.clone()))
         });
     }
     let order_result = {
@@ -771,17 +797,23 @@ pub fn circuit_to_schedule(
             node_memories: vec![],
             string: "".to_owned(),
         });
-        let mut mem = context.cache.max_tensor_elements;
-        while mem <= context.cache.fallback_total_numel && result.is_err() {
+        let mut mem = context.settings.max_memory;
+        while mem
+            <= context
+                .settings
+                .max_memory_fallback
+                .unwrap_or(context.settings.max_memory)
+            && result.is_err()
+        {
             if context.settings.scheduling_simplify {
                 let mut dag_simp_settings: DagSimpSettings = Default::default();
-                dag_simp_settings.acceptable_subdag_max = context.cache.max_single_tensor_numel;
+                dag_simp_settings.acceptable_subdag_max = context.settings.max_memory;
                 dag_simp_settings.mem_limit = mem;
                 dag_simp_settings.num_mem_chunks = context.settings.scheduling_num_mem_chunks;
                 dag_simp_settings.timeout = context.settings.scheduling_timeout;
                 dag_simp_settings.verbose = context.settings.verbose;
                 timed!(
-                    dag.simplify(&dag_simp_settings),
+                    dag.simplify(&dag_simp_settings)?,
                     10,
                     context.settings.verbose >= 2
                 );
