@@ -1,5 +1,6 @@
 use std::cmp::max;
 
+use anyhow::{bail, Result};
 use itertools::Itertools;
 use pyo3::prelude::*;
 use rr_util::{
@@ -9,13 +10,17 @@ use rr_util::{
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use smallvec::SmallVec as Sv;
 
-use crate::{scheduled_execution::SchedulingOOMError, scheduling_z3::schedule_dag_strategy_ints};
+use crate::{
+    scheduled_execution::{PySchedulingOOMManyError, SchedulingOOMError},
+    scheduling_z3::schedule_dag_strategy_ints,
+};
 pub struct DagSimpSettings {
     pub same_ratio: f64,
     pub tiny_size: usize,
     pub max_exhaustive_size: usize,
     pub max_z3_ints_size: usize,
     pub acceptable_subdag_max: usize,
+    pub exhaustive_settle_ns: usize,
     pub exhaustive_give_up_ns: usize,
     pub mem_limit: usize,
     pub num_mem_chunks: usize,
@@ -31,7 +36,8 @@ impl Default for DagSimpSettings {
             max_exhaustive_size: 60,
             max_z3_ints_size: 4000,
             acceptable_subdag_max: 100_000_000,
-            exhaustive_give_up_ns: 40_000,
+            exhaustive_settle_ns: 2_000_000,
+            exhaustive_give_up_ns: 400_000_000,
             verbose: 0,
             timeout: 5_000, // useless defaults, todo fix
             num_mem_chunks: 200,
@@ -46,10 +52,6 @@ impl Default for DagSimpSettings {
 pub struct Dag {
     pub children: Vec<Sv<[u32; 6]>>,
     pub parents: Vec<Sv<[u32; 6]>>,
-    /// the cost is the number of new elements computing the node entails allocating.
-    /// some nodes compute views into existing memory, such as some indexes and all Arrays
-    /// (as they're already allocated), so they have cost 0
-    /// if a node computes a new tensor, the cost is the product of the shape.
     pub node_costs: Vec<usize>,
     pub node_to_orig: HashMap<u32, Sv<[u32; 6]>>,
     pub node_hashes: Vec<HashBytes>,
@@ -71,6 +73,14 @@ impl Dag {
             .collect()
     }
 
+    pub fn cost(&self) -> usize {
+        self.node_to_orig
+            .iter()
+            .map(|x| self.node_costs[*x.0 as usize])
+            .max()
+            .unwrap_or(0)
+    }
+
     pub fn compute_schedule_top_ints(
         &self,
         verbose: usize,
@@ -148,36 +158,47 @@ impl Dag {
 
     // simplify all
 
-    pub fn simplify(&mut self, simp_settings: &DagSimpSettings) -> bool {
+    pub fn simplify(&mut self, simp_settings: &DagSimpSettings) -> Result<bool> {
         // println!(
         //     "simplifying to_orig {:?} children {:?}",
         //     &self.node_to_orig, &self.children
         // );
         // self.print_cytoscape_link();
         let mut result = false;
-        while self.simplify_pass(simp_settings, false) {
+        while self.simplify_pass(simp_settings, false)? {
             if simp_settings.verbose > 2 {
                 println!("did pass {}", self.node_to_orig.len());
                 self.print_cytoscape_link();
             }
             result = true;
         }
-        while self.simplify_pass(simp_settings, true) {
-            self.print_cytoscape_link();
+        while self.simplify_pass(simp_settings, true)? {
             if simp_settings.verbose > 2 {
+                self.print_cytoscape_link();
                 println!("did pass {}", self.node_to_orig.len());
             }
             result = true;
         }
-        self.simplify_pass(simp_settings, true);
-        result
+        if self.cost() > simp_settings.mem_limit {
+            bail!(SchedulingOOMError::Simp {
+                cyto: self.repr_cytoscape_link(),
+                origin: "post check".to_owned()
+            })
+        }
+        Ok(result)
     }
 
     // warning: I'm modifying the graph as I'm traversing it. edits must not mess with topo frontier
-    pub fn simplify_pass(&mut self, simp_settings: &DagSimpSettings, _do_isolated: bool) -> bool {
+    pub fn simplify_pass(
+        &mut self,
+        simp_settings: &DagSimpSettings,
+        do_isolated: bool,
+    ) -> Result<bool> {
         // simplify bottom_up
         let mut did_anything = false;
-
+        if do_isolated {
+            did_anything |= self.schedule_isolated(simp_settings)?;
+        }
         // we don't re-check whether leaves are still leaves, so
         // no rewrite can add children to an int which previously had no children
         let mut nodes_visited: HashSet<u32> = Default::default();
@@ -197,7 +218,7 @@ impl Dag {
             let simps: [(
                 &'static str,
                 fn(&mut Dag, u32, &DagSimpSettings) -> Option<Vec<u32>>,
-            ); 7] = [
+            ); 8] = [
                 and_namer!(Dag::try_inline_lone_sibling),
                 and_namer!(Dag::try_merge_simple_diamond),
                 and_namer!(Dag::try_merge_chain_close),
@@ -205,6 +226,7 @@ impl Dag {
                 and_namer!(Dag::try_merge_half_diamond),
                 and_namer!(Dag::try_elim_small_leaf),
                 and_namer!(Dag::try_merge_descending_peaks),
+                and_namer!(Dag::try_merge_multi_diamond_below),
             ];
             for (simp_name, simplification) in simps {
                 if let Some(the_parenties) = simplification(self, popped, simp_settings) {
@@ -244,9 +266,9 @@ impl Dag {
         if nodes_visited.len() < self.node_to_orig.len() {
             println!("didnt visit enough nodes {:?}", &self);
         }
-        did_anything |= self.schedule_isolated(simp_settings);
+
         // println!("did it {:?}", self.node_to_orig);
-        did_anything
+        Ok(did_anything)
     }
 
     // try simplification at point
@@ -304,6 +326,34 @@ impl Dag {
         None
     }
 
+    pub fn try_merge_multi_diamond_below(
+        &mut self,
+        node: u32,
+        _simp_settings: &DagSimpSettings,
+    ) -> Option<Vec<u32>> {
+        if self.children[node as usize].len() <= 1 {
+            return None;
+        }
+        if self.children[node as usize]
+            .iter()
+            .all(|&c| self.parents[c as usize].len() == 1 && self.children[c as usize].len() == 1)
+            && self.children[node as usize].iter().all(|&c| {
+                self.children[c as usize][0]
+                    == self.children[self.children[node as usize][0] as usize][0]
+            })
+        {
+            let mut new_cost = 0;
+            let intermediate = self.split_node_extra_below(node);
+            for c in self.children[intermediate as usize].clone() {
+                self.merge_add(intermediate, c, true);
+                new_cost += self.node_costs[c as usize];
+            }
+            self.node_costs[intermediate as usize] = new_cost;
+            return Some(Default::default());
+        }
+        None
+    }
+
     pub fn try_merge_chain_close(
         &mut self,
         node: u32,
@@ -546,13 +596,27 @@ impl Dag {
             .extend(self.node_to_orig.remove(&node).unwrap());
     }
 
+    pub fn seperate_node(&mut self, node: u32) {
+        assert!(self.children[node as usize].is_empty());
+        for parent in &self.parents[node as usize] {
+            self.children[*parent as usize] = self.children[*parent as usize]
+                .iter()
+                .filter(|x| **x != node)
+                .cloned()
+                .collect();
+        }
+        self.pre_root_nodes
+            .extend(self.node_to_orig.remove(&node).unwrap());
+    }
+
     // passes
 
     // schedule isolated pass
-    pub fn schedule_isolated(&mut self, settings: &DagSimpSettings) -> bool {
+    pub fn schedule_isolated(&mut self, settings: &DagSimpSettings) -> Result<bool> {
         let isolated = self.find_isolated(settings);
         // self.print_cytoscape_link();
         // println!("{:?} ", &isolated);
+        let mut did_anything = false;
         for iso in &isolated {
             let direct_only_sub = self.get_direct_only_sub(&iso.2);
             // direct_only_sub.print_cytoscape_link();
@@ -567,15 +631,19 @@ impl Dag {
             let order: Vec<(u32, usize)> = if iso.2.len() < settings.max_exhaustive_size
                 || iso.2.len() > settings.max_exhaustive_size
             {
-                timed!(
-                    direct_only_sub
-                        .schedule_exhaustive(required_first, settings)
-                        .iter()
-                        .map(|(k, v)| (direct_only_sub.node_to_orig[k][0], *v))
-                        .collect(),
+                let resy = timed!(
+                    direct_only_sub.schedule_exhaustive(required_first, settings),
                     10,
                     settings.verbose >= 2
-                )
+                );
+                if resy.is_err() && did_anything {
+                    return Ok(true);
+                }
+
+                resy?
+                    .iter()
+                    .map(|(k, v)| (direct_only_sub.node_to_orig[k][0], *v))
+                    .collect()
             } else {
                 let order = direct_only_sub
                     .compute_schedule_top_ints(
@@ -608,10 +676,11 @@ impl Dag {
             if settings.verbose > 2 {
                 println!("did isolated");
             }
+            did_anything = true;
             // println!("FINISHED ISOLATED SCHEDULE");
         }
         // self.print_cytoscape_link();
-        !isolated.is_empty()
+        Ok(!isolated.is_empty())
     }
 
     pub fn find_isolated(&self, _settings: &DagSimpSettings) -> Vec<(u32, u32, HashSet<u32>)> {
@@ -655,7 +724,7 @@ impl Dag {
                 }
             }
         }
-        result.sort_by(|a, b| a.2.len().cmp(&b.2.len()));
+        result.sort_unstable_by(|a, b| a.2.len().cmp(&b.2.len()));
         result
     }
 
@@ -663,10 +732,10 @@ impl Dag {
         &self,
         entry_point: Option<u32>,
         settings: &DagSimpSettings,
-    ) -> Vec<(u32, usize)> {
+    ) -> Result<Vec<(u32, usize)>> {
         let mut parents_left: Vec<i32> = self.parents.iter().map(|x| x.len() as i32).collect();
         let mut children_left: Vec<i32> = self.children.iter().map(|x| x.len() as i32).collect();
-        let mut best: (Vec<(u32, usize)>, usize) = (vec![], settings.mem_limit + 1);
+        let mut best: (Vec<(u32, usize)>, usize) = (vec![], settings.acceptable_subdag_max + 1);
         let mut order: Vec<(u32, u32, usize)> = Vec::with_capacity(self.node_costs.len());
 
         let mut options_stack: Vec<Sv<[u32; 16]>> = Vec::with_capacity(self.node_costs.len() + 2);
@@ -692,7 +761,16 @@ impl Dag {
                     .map(|x| self.node_costs[x as usize])
                     .sum();
                 order.push((choice, $i, cost));
-                options_stack.push(filter_to_idx(&children_left, |x| *x == 0).collect())
+                options_stack.push(filter_to_idx(&children_left, |x| *x == 0).collect());
+                // trying out sorting options, must do this before stack is used
+                let endel1 = options_stack.len() - 1;
+                options_stack[endel1].sort_by_cached_key(|&x| {
+                    self.parents[x as usize]
+                        .iter()
+                        .map(|&parent| children_left[parent as usize])
+                        .max()
+                        .unwrap_or(0)
+                })
             }};
         }
         macro_rules! pop {
@@ -730,17 +808,15 @@ impl Dag {
             }};
         }
         let start_instant = std::time::Instant::now();
-        loop {
+        for numiters in 0usize.. {
             // dbg!(&order, &options_stack, &parents_left, &children_left, &live);
             if order.len() == self.node_costs.len() {
                 best = (
                     order.iter().map(|(a, _b, c)| (*a, *c)).collect(),
                     *order.iter().map(|(_n, _i, c)| c).max().unwrap(),
                 );
-                if best.1 < settings.acceptable_subdag_max {
-                    if start_instant.elapsed().as_nanos() > settings.exhaustive_give_up_ns as u128 {
-                        break;
-                    }
+                if start_instant.elapsed().as_nanos() > settings.exhaustive_settle_ns as u128 {
+                    break;
                 }
                 backtrack!();
             } else {
@@ -752,8 +828,16 @@ impl Dag {
             if order.len() == 0 {
                 break;
             }
+            if numiters.trailing_zeros() > 8
+                && start_instant.elapsed().as_nanos() > settings.exhaustive_give_up_ns as u128
+            {
+                bail!(SchedulingOOMError::Simp {
+                    cyto: self.repr_cytoscape_link(),
+                    origin: "exhaustive timeout".to_owned()
+                })
+            }
         }
-        best.0
+        Ok(best.0)
     }
 
     pub fn replace_isolated_with_linear(&mut self, order: &Vec<(u32, usize)>) {
@@ -885,6 +969,9 @@ impl Dag {
 
     // misc
     pub fn print_cytoscape_link(&self) {
+        python_println!("{}", self.repr_cytoscape_link());
+    }
+    pub fn repr_cytoscape_link(&self) -> String {
         let not_excaped = format!(
             "{{\"graphName\":\"\",\"graphValue\":[{},{}]}}",
             self.node_to_orig
@@ -907,7 +994,7 @@ impl Dag {
                 .collect::<Vec<_>>()
                 .join(",")
         );
-        python_println!(
+        format!(
             "http://interp-tools.redwoodresearch.org/#/networkgraph/{}",
             url_escape::encode_fragment(&not_excaped)
         )
