@@ -4,7 +4,7 @@ use std::{
     iter::{once, zip},
 };
 
-use anyhow::{Error, Result};
+use anyhow::Result;
 use circuit_base::{
     clicolor,
     print::{color, last_child_arrows, CliColor, PrintOptions},
@@ -17,7 +17,7 @@ use rr_util::{
     python_println,
     util::{indent, indent_butfirst},
 };
-use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use rustc_hash::FxHashSet as HashSet;
 
 use crate::{
     iterative_matcher::{per_child, ChainItem, IterativeMatcher},
