@@ -25,7 +25,7 @@ use rr_util::{
     tu8v,
     util::{flip_op_result, HashBytes},
 };
-use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
+use rustc_hash::FxHashSet as HashSet;
 use smallvec::SmallVec as Sv;
 use thiserror::Error;
 
