@@ -16,6 +16,7 @@ use circuit_rewrites::{
     module_rewrite::{extract_rewrite_raw, module_remove_unused_inputs},
 };
 use get_update_node::{AnyFound, IterativeMatcher, IterativeMatcherRc, MatcherData};
+use itertools::multiunzip;
 use macro_rules_attribute::apply;
 use pyo3::{
     exceptions::{PyRuntimeError, PyValueError},
@@ -32,6 +33,7 @@ use rr_util::{
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
+use uuid::uuid;
 
 #[pyfunction(
     prefix_to_strip = "None",
@@ -306,7 +308,8 @@ pub fn fuse_modules_impl(
 #[derive(Debug, Clone)]
 struct NestedModuleItems {
     modules: Vec<Module>,
-    new_nodes: Vec<Vec<CircuitRc>>,
+    new_nodes_flat: Vec<Vec<CircuitRc>>,
+    new_nodes_base: Vec<Vec<CircuitRc>>,
     flat: Vec<(CircuitRc, ModuleArgSpec)>,
     flat_sym_to_arg: HashMap<Symbol, CircuitRc>,
     hash: HashBytes,
@@ -314,17 +317,24 @@ struct NestedModuleItems {
 
 impl Default for NestedModuleItems {
     fn default() -> Self {
-        Self::new(vec![], vec![]).unwrap()
+        Self::new(vec![], vec![], vec![]).unwrap()
     }
 }
 
 impl NestedModuleItems {
-    fn new(modules: Vec<Module>, new_nodes: Vec<Vec<CircuitRc>>) -> Result<Self> {
-        for (self_m, new) in modules.iter().zip(&new_nodes) {
+    fn new(
+        modules: Vec<Module>,
+        new_nodes_flat: Vec<Vec<CircuitRc>>,
+        new_nodes_base: Vec<Vec<CircuitRc>>,
+    ) -> Result<Self> {
+        assert_eq!(modules.len(), new_nodes_flat.len());
+        assert_eq!(modules.len(), new_nodes_base.len());
+        for ((self_m, new), base) in modules.iter().zip(&new_nodes_flat).zip(&new_nodes_base) {
             assert_eq!(self_m.nodes.len(), new.len());
+            assert_eq!(self_m.nodes.len(), base.len());
         }
 
-        let flat = fuse_modules_impl(&modules, &new_nodes)?;
+        let flat = fuse_modules_impl(&modules, &new_nodes_flat)?;
         let flat_sym_to_arg = flat
             .iter()
             .map(|(circ, arg_spec)| (arg_spec.symbol.clone(), circ.clone()))
@@ -332,28 +342,42 @@ impl NestedModuleItems {
 
         let mut hasher = blake3::Hasher::new();
 
-        for (m, this_new_nodes) in modules.iter().zip(&new_nodes) {
-            for n in this_new_nodes {
+        for ((m, this_new_nodes_flat), this_new_nodes_base) in
+            modules.iter().zip(&new_nodes_flat).zip(&new_nodes_base)
+        {
+            for n in this_new_nodes_flat {
                 hasher.update(&n.hash());
             }
+            for n in this_new_nodes_base {
+                hasher.update(&n.hash());
+            }
+            hasher.update(uuid!("a5685fe8-00a8-4680-bf64-a5802b36fd7d").as_bytes());
             hasher.update(&m.hash());
         }
 
         Ok(Self {
             modules,
-            new_nodes,
+            new_nodes_flat,
+            new_nodes_base,
             flat,
             flat_sym_to_arg,
             hash: hasher.finalize().into(),
         })
     }
 
-    fn push(&self, m: Module, this_new_nodes: Vec<CircuitRc>) -> Result<Self> {
+    fn push(
+        &self,
+        m: Module,
+        this_new_nodes_flat: Vec<CircuitRc>,
+        this_new_nodes_base: Vec<CircuitRc>,
+    ) -> Result<Self> {
         let mut modules = self.modules.clone();
-        let mut new_nodes = self.new_nodes.clone();
+        let mut new_nodes_flat = self.new_nodes_flat.clone();
+        let mut new_nodes_base = self.new_nodes_base.clone();
         modules.push(m);
-        new_nodes.push(this_new_nodes);
-        Self::new(modules, new_nodes)
+        new_nodes_flat.push(this_new_nodes_flat);
+        new_nodes_base.push(this_new_nodes_base);
+        Self::new(modules, new_nodes_flat, new_nodes_base)
     }
 }
 
@@ -413,6 +437,8 @@ pub struct ModulePusher {
     #[pyo3(get)]
     remove_unused_inputs: bool,
     #[pyo3(get)]
+    add_suffix_on_remove_unused: bool,
+    #[pyo3(get)]
     elim_no_input_modules: bool,
     #[pyo3(get)]
     bind_encountered_symbols: bool,
@@ -437,6 +463,7 @@ impl Default for ModulePusher {
         Self {
             flatten_modules: true,
             remove_unused_inputs: true,
+            add_suffix_on_remove_unused: false,
             elim_no_input_modules: true,
             bind_encountered_symbols: true,
             namer: Default::default(),
@@ -452,6 +479,7 @@ impl ModulePusher {
     #[args(
         flatten_modules = "Self::default().flatten_modules",
         remove_unused_inputs = "Self::default().remove_unused_inputs",
+        add_suffix_on_remove_unused = "Self::default().add_suffix_on_remove_unused",
         elim_no_input_modules = "Self::default().elim_no_input_modules",
         bind_encountered_symbols = "Self::default().bind_encountered_symbols",
         namer = "Self::default().namer"
@@ -459,6 +487,7 @@ impl ModulePusher {
     pub fn new(
         flatten_modules: bool,
         remove_unused_inputs: bool,
+        add_suffix_on_remove_unused: bool,
         elim_no_input_modules: bool,
         bind_encountered_symbols: bool,
         namer: NestedModuleNamer,
@@ -466,6 +495,7 @@ impl ModulePusher {
         Self {
             flatten_modules,
             remove_unused_inputs,
+            add_suffix_on_remove_unused,
             elim_no_input_modules,
             bind_encountered_symbols,
             namer,
@@ -634,12 +664,34 @@ impl ModulePusher {
             assert_eq!(node_get_skip.len(), m.nodes.len());
 
             // intentionally don't use extra_replacements per child
-            let new_nodes = m
+            let new_nodes_flat = m
+                .nodes
+                .clone()
+                .into_iter()
+                .zip(node_get_skip.clone())
+                .map(|(c, (get, skip))| rec(self_, c, get, skip, items, extra_replacements))
+                .collect::<Result<Vec<_>>>()?;
+
+            let new_nodes_base = m
                 .nodes
                 .clone()
                 .into_iter()
                 .zip(node_get_skip)
-                .map(|(c, (get, skip))| rec(self_, c, get, skip, items, extra_replacements))
+                // don't use rec to avoid adding nodes to out_set
+                .map(|(c, (get, skip))| {
+                    let ndim = c.ndim();
+                    let out = self_
+                        .push_down_modules_rec(
+                            c,
+                            get,
+                            skip,
+                            is_get,
+                            &Default::default(), // no items (only difference from recursive call above)
+                            extra_replacements,
+                        )?.0;
+                    assert_eq!(out.ndim(), ndim);
+                    Ok(out)
+                })
                 .collect::<Result<Vec<_>>>()?;
 
             rec(
@@ -647,7 +699,7 @@ impl ModulePusher {
                 m.spec.circuit.clone(),
                 spec_circuit_get,
                 spec_circuit_skip_module,
-                &items.push(m.clone(), new_nodes)?,
+                &items.push(m.clone(), new_nodes_flat, new_nodes_base)?,
                 extra_replacements,
             )?
         } else {
@@ -776,14 +828,21 @@ impl ModulePusher {
                 // if we have extra dims, we have to pull the symbols and resub to sync up batching
                 let rev_nested: Vec<_> = items
                     .modules
-                    .clone()
-                    .into_iter()
+                    .iter()
+                    .zip(&items.new_nodes_base)
                     .rev()
-                    .scan(0, |prev_dims_covered, m| {
+                    .scan(0, |prev_dims_covered, (m, new_nodes)| {
                         let extra_dims_here =
                             rep_expand_extra_dims.saturating_sub(*prev_dims_covered);
                         *prev_dims_covered += m.aligned_batch_shape().len();
-                        Some(update_flat(m.arg_items(), extra_dims_here))
+                        Some(update_flat(
+                            new_nodes
+                                .clone()
+                                .into_iter()
+                                .zip(m.spec.arg_specs.clone())
+                                .collect(),
+                            extra_dims_here,
+                        ))
                     })
                     .collect();
                 let new_finished_circuit = map_replace_expand(
@@ -796,10 +855,16 @@ impl ModulePusher {
                     finished_circuit,
                     items
                         .modules
-                        .clone()
-                        .into_iter()
+                        .iter()
+                        .zip(&items.new_nodes_base)
                         .rev()
-                        .map(|m| m.arg_items())
+                        .map(|(m, new_nodes)| {
+                            new_nodes
+                                .clone()
+                                .into_iter()
+                                .zip(m.spec.arg_specs.clone())
+                                .collect()
+                        })
                         .collect(),
                 )
             };
@@ -822,7 +887,11 @@ impl ModulePusher {
             .context("namer failed in push down module")?;
         let m_out = Module::new(nodes, ModuleSpec { circuit, arg_specs }, name.clone());
         if self.remove_unused_inputs {
-            module_remove_unused_inputs(&m_out, self.elim_no_input_modules)
+            module_remove_unused_inputs(
+                &m_out,
+                self.add_suffix_on_remove_unused,
+                self.elim_no_input_modules,
+            )
         } else {
             Ok(m_out.rc())
         }
@@ -908,7 +977,7 @@ impl_eq_by_big_hash!(SymbolSetRc);
 
 struct SymbolExtractor {
     cache: cached::UnboundCache<
-        (HashBytes, HashBytes),
+        (HashBytes, HashBytes, IterativeMatcherRc),
         (CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>),
     >,
     conform_batch_if_needed: bool,
@@ -926,14 +995,26 @@ impl Default for SymbolExtractor {
 impl SymbolExtractor {
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key((circ.info().hash, symbols.hash()))]
+    #[key((circ.info().hash, symbols.hash(), traversal.clone()))]
     #[use_try]
     #[cache_expr(cache)]
     fn extract_symbols_rec(
         &mut self,
         circ: CircuitRc,
         symbols: &SymbolSetRc,
+        traversal: IterativeMatcherRc,
     ) -> Result<(CircuitRc, BTreeMap<Symbol, (CircuitRc, ModuleArgSpec)>)> {
+        let updated = traversal
+            .match_iterate(circ.clone())?
+            .unwrap_or_same(traversal)
+            .0;
+
+        if updated.all_finished() {
+            return Ok((circ, Default::default()));
+        }
+
+        let traversal_per_child = updated.per_child_with_term(circ.num_children());
+
         let mut out_syms = BTreeMap::default();
         let add_sym_circ = |out_syms: &mut BTreeMap<_, (CircuitRc, ModuleArgSpec)>,
                             arg_spec: ModuleArgSpec,
@@ -961,12 +1042,8 @@ impl SymbolExtractor {
             Ok(())
         };
 
-        // do before we extract
-        // TODO: add flag for nested extract?
-        // let per_child_syms = symbols.per_child(&circ);
-
         let mut freshly_bound_syms = HashSet::default(); // so we extract outer -> inner (left -> right)
-        let circ = if let Some(m) = circ.as_module() {
+        let (circ, traversal_per_child) = if let Some(m) = circ.as_module() {
             let new_mod = if self_.conform_batch_if_needed {
                 let batch = m
                     .nodes
@@ -984,15 +1061,18 @@ impl SymbolExtractor {
                 None
             };
             let new_m = new_mod.as_ref().unwrap_or(m);
-            let (nodes, arg_specs): (Vec<_>, Vec<_>) = new_m
+            let mut traversals_iter = traversal_per_child.into_iter();
+            let spec_traversal = traversals_iter.next().unwrap();
+            let iter = new_m
                 .arg_items()
                 .into_iter()
                 .zip(&m.spec.arg_specs)
-                .filter_map(|((node, arg_spec), orig_arg_spec)| {
+                .zip(traversals_iter)
+                .filter_map(|(((node, arg_spec), orig_arg_spec), traversal)| {
                     if !symbols.contains(&orig_arg_spec.symbol)
                         || freshly_bound_syms.contains(&orig_arg_spec.symbol)
                     {
-                        return Some(Ok((node, arg_spec)));
+                        return Some(Ok((node, arg_spec, traversal)));
                     }
                     if node.ndim() > arg_spec.symbol.ndim() {
                         assert!(!self_.conform_batch_if_needed); // this should have been handled
@@ -1010,10 +1090,9 @@ impl SymbolExtractor {
                     freshly_bound_syms.insert(orig_arg_spec.symbol.clone());
                     None
                 })
-                .collect::<Result<Vec<_>>>()?
-                .into_iter()
-                .unzip();
-            if nodes.len() == m.nodes.len() {
+                .collect::<Result<Vec<_>>>()?;
+            let (nodes, arg_specs, traversal_per_node): (Vec<_>, Vec<_>, Vec<_>) = multiunzip(iter);
+            let circ = if nodes.len() == m.nodes.len() {
                 m.crc()
             } else {
                 Module::nrc(
@@ -1024,9 +1103,13 @@ impl SymbolExtractor {
                     },
                     m.name().map(|s| format!("{} extracted", s)),
                 )
-            }
+            };
+            let traversal_per_child = iter::once(spec_traversal)
+                .chain(traversal_per_node)
+                .collect();
+            (circ, traversal_per_child)
         } else {
-            circ
+            (circ, traversal_per_child)
         };
 
         let per_child_syms = if !freshly_bound_syms.is_empty() {
@@ -1043,8 +1126,11 @@ impl SymbolExtractor {
             vec![symbols.clone(); circ.num_children()]
         };
 
+        assert_eq!(per_child_syms.len(), circ.num_children());
+        assert_eq!(traversal_per_child.len(), circ.num_children());
         let new_circ = circ.map_children_enumerate(|i, c| {
-            let (new, map) = self_.extract_symbols_rec(c, &per_child_syms[i])?;
+            let (new, map) =
+                self_.extract_symbols_rec(c, &per_child_syms[i], traversal_per_child[i].clone())?;
             for (_, (circ, arg_spec)) in map {
                 add_sym_circ(&mut out_syms, arg_spec, circ)?;
             }
@@ -1074,19 +1160,24 @@ impl SymbolExtractor {
     }
 }
 
-#[pyfunction(use_elim_no_input_modules = "true", conform_batch_if_needed = "false")]
+#[pyfunction(
+    use_elim_no_input_modules = "true",
+    conform_batch_if_needed = "false",
+    traversal = "IterativeMatcher::noop_traversal().rc()"
+)]
 pub fn extract_symbols(
     circuit: CircuitRc,
     symbols: SymbolSetRc,
     use_elim_no_input_modules: bool,
     conform_batch_if_needed: bool,
+    traversal: IterativeMatcherRc,
 ) -> Result<Module> {
     let name = circuit.name_cloned();
     let (extracted_circ, extracted_syms) = SymbolExtractor {
         conform_batch_if_needed,
         ..Default::default()
     }
-    .extract_symbols_rec(circuit, &symbols)?;
+    .extract_symbols_rec(circuit, &symbols, traversal)?;
 
     let extracted_circ = if use_elim_no_input_modules {
         SimpFnSubset::none()
@@ -1109,12 +1200,17 @@ pub fn extract_symbols(
     ))
 }
 
-#[pyfunction(use_elim_no_input_modules = "true", conform_batch_if_needed = "false")]
+#[pyfunction(
+    use_elim_no_input_modules = "true",
+    conform_batch_if_needed = "false",
+    traversal = "IterativeMatcher::noop_traversal().rc()"
+)]
 pub fn extract_symbols_get(
     circuit: CircuitRc,
     get: IterativeMatcherRc,
     use_elim_no_input_modules: bool,
     conform_batch_if_needed: bool,
+    traversal: IterativeMatcherRc,
 ) -> Result<Module> {
     let syms = get
         .get(circuit.clone(), false)
@@ -1135,6 +1231,7 @@ pub fn extract_symbols_get(
         SymbolSetRc::new(syms),
         use_elim_no_input_modules,
         conform_batch_if_needed,
+        traversal,
     )
 }
 
