@@ -1,4 +1,4 @@
-use std::fmt;
+use std::{fmt, ops::Deref};
 
 use anyhow::{Context, Result};
 use macro_rules_attribute::apply;
@@ -8,7 +8,6 @@ use pyo3::{
     pyclass::CompareOp,
     types::{IntoPyDict, PyTuple},
 };
-use rustc_hash::FxHashMap as HashMap;
 
 use crate::{
     tensor_util::{Shape, TorchDeviceDtype},
@@ -478,6 +477,45 @@ impl IntoPy<PyObject> for PyEinsumAxes {
     }
 }
 
+#[derive(Debug, Clone)]
+pub struct PyCallable(PyObject);
+
+impl fmt::Display for PyCallable {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        self.0.fmt(f)
+    }
+}
+
+impl Deref for PyCallable {
+    type Target = PyObject;
+    fn deref(&self) -> &Self::Target {
+        &self.0
+    }
+}
+
+impl PyCallable {
+    pub fn new(callable: PyObject) -> PyResult<Self> {
+        if !Python::with_gil(|py| callable.as_ref(py).is_callable()) {
+            return Err(PyErr::new::<pyo3::exceptions::PyTypeError, _>(
+                "Value isn't callable!",
+            ));
+        }
+        Ok(Self(callable))
+    }
+}
+
+impl<'source> FromPyObject<'source> for PyCallable {
+    fn extract(ob: &'source PyAny) -> PyResult<Self> {
+        Self::new(ob.into())
+    }
+}
+
+impl IntoPy<PyObject> for PyCallable {
+    fn into_py(self, _py: Python<'_>) -> PyObject {
+        self.0
+    }
+}
+
 pub fn use_rust_comp<T: PartialOrd>(l: &T, r: &T, comp_op: CompareOp) -> bool {
     match comp_op {
         CompareOp::Lt => l < r,
