@@ -14,7 +14,7 @@ use pyo3::{
     exceptions::PyValueError,
     ffi,
     prelude::*,
-    pyclass::{self, CompareOp},
+    pyclass::CompareOp,
     pymethods,
     types::{IntoPyDict, PySlice, PyTuple},
     AsPyPointer, FromPyObject, IntoPy, PyAny, PyErr, PyObject, PyResult, Python, ToPyObject,
